@@ -17,26 +17,35 @@ from openstack import exceptions
 from openstack.image import _base_proxy
 from openstack.image.v2 import image as _image
 from openstack.image.v2 import member as _member
+from openstack.image.v2 import metadef_namespace as _metadef_namespace
+from openstack.image.v2 import metadef_schema as _metadef_schema
 from openstack.image.v2 import schema as _schema
-from openstack.image.v2 import task as _task
 from openstack.image.v2 import service_info as _si
+from openstack.image.v2 import task as _task
 from openstack import resource
 from openstack import utils
 
 # Rackspace returns this for intermittent import errors
 _IMAGE_ERROR_396 = "Image cannot be imported. Error code: '396'"
 _INT_PROPERTIES = ('min_disk', 'min_ram', 'size', 'virtual_size')
+_RAW_PROPERTIES = ('is_protected', 'tags')
 
 
 class Proxy(_base_proxy.BaseImageProxy):
 
+    # ====== IMAGES ======
     def _create_image(self, **kwargs):
         """Create image resource from attributes
         """
         return self._create(_image.Image, **kwargs)
 
-    def import_image(self, image, method='glance-direct', uri=None,
-                     store=None):
+    def import_image(
+        self, image, method='glance-direct', uri=None,
+        store=None,
+        stores=None,
+        all_stores=None,
+        all_stores_must_succeed=None,
+    ):
         """Import data to an existing image
 
         Interoperable image import process are introduced in the Image API
@@ -44,24 +53,57 @@ class Proxy(_base_proxy.BaseImageProxy):
         Image Service download it by itself without sending binary data at
         image creation.
 
-        :param image: The value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
-        :param method: Method to use for importing the image.
-                       A valid value is glance-direct or web-download.
-        :param uri: Required only if using the web-download import method.
-                    This url is where the data is made available to the Image
-                    service.
-        :param store: Used when enabled_backends is activated in glance
-                      The value can be the id of a store or a
-                      :class:`~openstack.image.v2.service_info.Store`
-                      instance.
+        :param image:
+            The value can be the ID of a image or a
+            :class:`~openstack.image.v2.image.Image` instance.
+        :param method:
+            Method to use for importing the image.
+            A valid value is glance-direct or web-download.
+        :param uri:
+            Required only if using the web-download import method.
+            This url is where the data is made available to the Image
+            service.
+        :param store:
+            Used when enabled_backends is activated in glance. The value
+            can be the id of a store or a
+            :class:`~openstack.image.v2.service_info.Store` instance.
+        :param stores:
+            List of stores to be used when enabled_backends is activated
+            in glance. List values can be the id of a store or a
+            :class:`~openstack.image.v2.service_info.Store` instance.
+        :param all_stores:
+            Upload to all available stores. Mutually exclusive with
+            ``store`` and ``stores``.
+        :param all_stores_must_succeed:
+            When set to True, if an error occurs during the upload in at
+            least one store, the worfklow fails, the data is deleted
+            from stores where copying is done (not staging), and the
+            state of the image is unchanged. When set to False, the
+            workflow will fail (data deleted from stores, …) only if the
+            import fails on all stores specified by the user. In case of
+            a partial success, the locations added to the image will be
+            the stores where the data has been correctly uploaded.
+            Default is True.
 
         :returns: None
         """
         image = self._get_resource(_image.Image, image)
+        if all_stores and (store or stores):
+            raise exceptions.InvalidRequest(
+                "all_stores is mutually exclusive with"
+                " store and stores")
         if store is not None:
+            if stores:
+                raise exceptions.InvalidRequest(
+                    "store and stores are mutually exclusive")
             store = self._get_resource(_si.Store, store)
 
+        stores = stores or []
+        new_stores = []
+        for s in stores:
+            new_stores.append(self._get_resource(_si.Store, s))
+        stores = new_stores
+
         # as for the standard image upload function, container_format and
         # disk_format are required for using image import process
         if not all([image.container_format, image.disk_format]):
@@ -69,7 +111,13 @@ class Proxy(_base_proxy.BaseImageProxy):
                 "Both container_format and disk_format are required for"
                 " importing an image")
 
-        image.import_image(self, method=method, uri=uri, store=store)
+        image.import_image(
+            self, method=method, uri=uri,
+            store=store,
+            stores=stores,
+            all_stores=all_stores,
+            all_stores_must_succeed=all_stores_must_succeed,
+        )
 
     def stage_image(self, image, filename=None, data=None):
         """Stage binary image data
@@ -147,9 +195,15 @@ class Proxy(_base_proxy.BaseImageProxy):
 
         return img
 
-    def _upload_image(self, name, filename=None, meta=None,
-                      wait=False, timeout=None, validate_checksum=True,
-                      **kwargs):
+    def _upload_image(
+        self, name, filename=None, data=None, meta=None,
+        wait=False, timeout=None, validate_checksum=True,
+        use_import=False,
+        stores=None,
+        all_stores=None,
+        all_stores_must_succeed=None,
+        **kwargs
+    ):
         # We can never have nice things. Glance v1 took "is_public" as a
         # boolean. Glance v2 takes "visibility". If the user gives us
         # is_public, we know what they mean. If they give us visibility, they
@@ -164,13 +218,23 @@ class Proxy(_base_proxy.BaseImageProxy):
         try:
             # This makes me want to die inside
             if self._connection.image_api_use_tasks:
+                if use_import:
+                    raise exceptions.SDKException(
+                        "The Glance Task API and Import API are"
+                        " mutually exclusive. Either disable"
+                        " image_api_use_tasks in config, or"
+                        " do not request using import")
                 return self._upload_image_task(
-                    name, filename, meta=meta,
+                    name, filename, data=data, meta=meta,
                     wait=wait, timeout=timeout, **kwargs)
             else:
                 return self._upload_image_put(
-                    name, filename, meta=meta,
+                    name, filename, data=data, meta=meta,
                     validate_checksum=validate_checksum,
+                    use_import=use_import,
+                    stores=stores,
+                    all_stores=all_stores,
+                    all_stores_must_succeed=all_stores_must_succeed,
                     **kwargs)
         except exceptions.SDKException:
             self.log.debug("Image creation failed", exc_info=True)
@@ -184,7 +248,7 @@ class Proxy(_base_proxy.BaseImageProxy):
         for k, v in iter(properties.items()):
             if k in _INT_PROPERTIES:
                 ret[k] = int(v)
-            elif k == 'protected':
+            elif k in _RAW_PROPERTIES:
                 ret[k] = v
             else:
                 if v is None:
@@ -195,8 +259,17 @@ class Proxy(_base_proxy.BaseImageProxy):
         return ret
 
     def _upload_image_put(
-            self, name, filename, meta, validate_checksum, **image_kwargs):
-        image_data = open(filename, 'rb')
+        self, name, filename, data, meta,
+        validate_checksum, use_import=False,
+        stores=None,
+        all_stores=None,
+        all_stores_must_succeed=None,
+        **image_kwargs,
+    ):
+        if filename and not data:
+            image_data = open(filename, 'rb')
+        else:
+            image_data = data
 
         properties = image_kwargs.pop('properties', {})
 
@@ -206,10 +279,25 @@ class Proxy(_base_proxy.BaseImageProxy):
         image = self._create(_image.Image, **image_kwargs)
 
         image.data = image_data
+        supports_import = (
+            image.image_import_methods
+            and 'glance-direct' in image.image_import_methods
+        )
+        if stores or all_stores or all_stores_must_succeed:
+            use_import = True
+        if use_import and not supports_import:
+            raise exceptions.SDKException(
+                "Importing image was requested but the cloud does not"
+                " support the image import method.")
 
         try:
-            response = image.upload(self)
-            exceptions.raise_from_response(response)
+            if not use_import:
+                response = image.upload(self)
+                exceptions.raise_from_response(response)
+            if use_import:
+                image.stage(self)
+                image.import_image(self)
+
             # image_kwargs are flat here
             md5 = image_kwargs.get(self._IMAGE_MD5_KEY)
             sha256 = image_kwargs.get(self._IMAGE_SHA256_KEY)
@@ -231,7 +319,7 @@ class Proxy(_base_proxy.BaseImageProxy):
         return image
 
     def _upload_image_task(
-            self, name, filename,
+            self, name, filename, data,
             wait, timeout, meta, **image_kwargs):
 
         if not self._connection.has_service('object-store'):
@@ -250,8 +338,10 @@ class Proxy(_base_proxy.BaseImageProxy):
         self._connection.create_object(
             container, name, filename,
             md5=md5, sha256=sha256,
+            data=data,
             metadata={self._connection._OBJECT_AUTOCREATE_KEY: 'true'},
-            **{'content-type': 'application/octet-stream'})
+            **{'content-type': 'application/octet-stream',
+               'x-delete-after': str(24 * 60 * 60)})
         # TODO(mordred): Can we do something similar to what nodepool does
         # using glance properties to not delete then upload but instead make a
         # new "good" image and then mark the old one as "bad"
@@ -332,22 +422,22 @@ class Proxy(_base_proxy.BaseImageProxy):
         :ref:`download_image-stream-true`.
 
         :param image: The value can be either the ID of an image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
+            :class:`~openstack.image.v2.image.Image` instance.
 
         :param bool stream: When ``True``, return a :class:`requests.Response`
-                            instance allowing you to iterate over the
-                            response data stream instead of storing its entire
-                            contents in memory. See
-                            :meth:`requests.Response.iter_content` for more
-                            details. *NOTE*: If you do not consume
-                            the entirety of the response you must explicitly
-                            call :meth:`requests.Response.close` or otherwise
-                            risk inefficiencies with the ``requests``
-                            library's handling of connections.
+            instance allowing you to iterate over the
+            response data stream instead of storing its entire
+            contents in memory. See
+            :meth:`requests.Response.iter_content` for more
+            details. *NOTE*: If you do not consume
+            the entirety of the response you must explicitly
+            call :meth:`requests.Response.close` or otherwise
+            risk inefficiencies with the ``requests``
+            library's handling of connections.
 
 
-                            When ``False``, return the entire
-                            contents of the response.
+            When ``False``, return the entire
+            contents of the response.
         :param output: Either a file object or a path to store data into.
         :param int chunk_size: size in bytes to read from the wire and buffer
             at one time. Defaults to 1024
@@ -367,12 +457,12 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Delete an image
 
         :param image: The value can be either the ID of an image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
+            :class:`~openstack.image.v2.image.Image` instance.
         :param bool ignore_missing: When set to ``False``
-                    :class:`~openstack.exceptions.ResourceNotFound` will be
-                    raised when the image does not exist.
-                    When set to ``True``, no exception will be set when
-                    attempting to delete a nonexistent image.
+            :class:`~openstack.exceptions.ResourceNotFound` will be
+            raised when the image does not exist.
+            When set to ``True``, no exception will be set when
+            attempting to delete a nonexistent image.
 
         :returns: ``None``
         """
@@ -383,10 +473,10 @@ class Proxy(_base_proxy.BaseImageProxy):
 
         :param name_or_id: The name or ID of a image.
         :param bool ignore_missing: When set to ``False``
-                    :class:`~openstack.exceptions.ResourceNotFound` will be
-                    raised when the resource does not exist.
-                    When set to ``True``, None will be returned when
-                    attempting to find a nonexistent resource.
+            :class:`~openstack.exceptions.ResourceNotFound` will be
+            raised when the resource does not exist.
+            When set to ``True``, None will be returned when
+            attempting to find a nonexistent resource.
         :returns: One :class:`~openstack.image.v2.image.Image` or None
         """
         return self._find(_image.Image, name_or_id,
@@ -396,11 +486,11 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Get a single image
 
         :param image: The value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
+            :class:`~openstack.image.v2.image.Image` instance.
 
         :returns: One :class:`~openstack.image.v2.image.Image`
         :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
+            when no resource can be found.
         """
         return self._get(_image.Image, image)
 
@@ -408,7 +498,7 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Return a generator of images
 
         :param kwargs query: Optional query parameters to be sent to limit
-                                 the resources being returned.
+            the resources being returned.
 
         :returns: A generator of image objects
         :rtype: :class:`~openstack.image.v2.image.Image`
@@ -419,9 +509,9 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Update a image
 
         :param image: Either the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
-        :attrs kwargs: The attributes to update on the image represented
-                       by ``value``.
+            :class:`~openstack.image.v2.image.Image` instance.
+        :param attrs: The attributes to update on the image represented
+            by ``image``.
 
         :returns: The updated image
         :rtype: :class:`~openstack.image.v2.image.Image`
@@ -432,7 +522,7 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Deactivate an image
 
         :param image: Either the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
+            :class:`~openstack.image.v2.image.Image` instance.
 
         :returns: None
         """
@@ -443,7 +533,7 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Deactivate an image
 
         :param image: Either the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
+            :class:`~openstack.image.v2.image.Image` instance.
 
         :returns: None
         """
@@ -454,8 +544,8 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Add a tag to an image
 
         :param image: The value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance
-                      that the member will be created for.
+            :class:`~openstack.image.v2.image.Image` instance
+            that the member will be created for.
         :param str tag: The tag to be added
 
         :returns: None
@@ -467,8 +557,8 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Remove a tag to an image
 
         :param image: The value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance
-                      that the member will be created for.
+            :class:`~openstack.image.v2.image.Image` instance
+            that the member will be created for.
         :param str tag: The tag to be removed
 
         :returns: None
@@ -476,15 +566,16 @@ class Proxy(_base_proxy.BaseImageProxy):
         image = self._get_resource(_image.Image, image)
         image.remove_tag(self, tag)
 
+    # ====== IMAGE MEMBERS ======
     def add_member(self, image, **attrs):
         """Create a new member from attributes
 
         :param image: The value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance
-                      that the member will be created for.
+            :class:`~openstack.image.v2.image.Image` instance
+            that the member will be created for.
         :param dict attrs: Keyword arguments which will be used to create
-                           a :class:`~openstack.image.v2.member.Member`,
-                           comprised of the properties on the Member class.
+            a :class:`~openstack.image.v2.member.Member`,
+            comprised of the properties on the Member class.
 
         :returns: The results of member creation
         :rtype: :class:`~openstack.image.v2.member.Member`
@@ -492,16 +583,18 @@ class Proxy(_base_proxy.BaseImageProxy):
         image_id = resource.Resource._get_id(image)
         return self._create(_member.Member, image_id=image_id, **attrs)
 
-    def remove_member(self, member, image, ignore_missing=True):
+    def remove_member(self, member, image=None, ignore_missing=True):
         """Delete a member
 
         :param member: The value can be either the ID of a member or a
-                       :class:`~openstack.image.v2.member.Member` instance.
+            :class:`~openstack.image.v2.member.Member` instance.
+        :param image: The value can be either the ID of an image or a
+            :class:`~openstack.image.v2.image.Image` instance that the member
+            is part of. This is required if ``member`` is an ID.
         :param bool ignore_missing: When set to ``False``
-                    :class:`~openstack.exceptions.ResourceNotFound` will be
-                    raised when the member does not exist.
-                    When set to ``True``, no exception will be set when
-                    attempting to delete a nonexistent member.
+            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
+            the member does not exist. When set to ``True``, no exception will
+            be set when attempting to delete a nonexistent member.
 
         :returns: ``None``
         """
@@ -515,13 +608,13 @@ class Proxy(_base_proxy.BaseImageProxy):
 
         :param name_or_id: The name or ID of a member.
         :param image: This is the image that the member belongs to,
-                      the value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
+            the value can be the ID of a image or a
+            :class:`~openstack.image.v2.image.Image` instance.
         :param bool ignore_missing: When set to ``False``
-                    :class:`~openstack.exceptions.ResourceNotFound` will be
-                    raised when the resource does not exist.
-                    When set to ``True``, None will be returned when
-                    attempting to find a nonexistent resource.
+            :class:`~openstack.exceptions.ResourceNotFound` will be
+            raised when the resource does not exist.
+            When set to ``True``, None will be returned when
+            attempting to find a nonexistent resource.
         :returns: One :class:`~openstack.image.v2.member.Member` or None
         """
         image_id = resource.Resource._get_id(image)
@@ -532,25 +625,27 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Get a single member on an image
 
         :param member: The value can be the ID of a member or a
-                       :class:`~openstack.image.v2.member.Member` instance.
+            :class:`~openstack.image.v2.member.Member` instance.
         :param image: This is the image that the member belongs to.
-                      The value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
+            The value can be the ID of a image or a
+            :class:`~openstack.image.v2.image.Image` instance.
         :returns: One :class:`~openstack.image.v2.member.Member`
         :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
+            when no resource can be found.
         """
         member_id = resource.Resource._get_id(member)
         image_id = resource.Resource._get_id(image)
         return self._get(_member.Member, member_id=member_id,
                          image_id=image_id)
 
-    def members(self, image):
+    def members(self, image, **query):
         """Return a generator of members
 
         :param image: This is the image that the member belongs to,
-                      the value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
+            the value can be the ID of a image or a
+            :class:`~openstack.image.v2.image.Image` instance.
+        :param kwargs query: Optional query parameters to be sent to limit
+            the resources being returned.
 
         :returns: A generator of member objects
         :rtype: :class:`~openstack.image.v2.member.Member`
@@ -562,12 +657,12 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Update the member of an image
 
         :param member: Either the ID of a member or a
-                       :class:`~openstack.image.v2.member.Member` instance.
+            :class:`~openstack.image.v2.member.Member` instance.
         :param image: This is the image that the member belongs to.
-                      The value can be the ID of a image or a
-                      :class:`~openstack.image.v2.image.Image` instance.
-        :attrs kwargs: The attributes to update on the member represented
-                       by ``value``.
+            The value can be the ID of a image or a
+            :class:`~openstack.image.v2.image.Image` instance.
+        :param attrs: The attributes to update on the member represented
+            by ``member``.
 
         :returns: The updated member
         :rtype: :class:`~openstack.image.v2.member.Member`
@@ -577,12 +672,98 @@ class Proxy(_base_proxy.BaseImageProxy):
         return self._update(_member.Member, member_id=member_id,
                             image_id=image_id, **attrs)
 
+    # ====== METADEF NAMESPACES ======
+    def create_metadef_namespace(self, **attrs):
+        """Create a new metadef namespace from attributes
+
+        :param dict attrs: Keyword arguments which will be used to create
+            a :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
+            comprised of the properties on the MetadefNamespace class.
+
+        :returns: The results of metadef namespace creation
+        :rtype: :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
+        """
+        return self._create(_metadef_namespace.MetadefNamespace, **attrs)
+
+    def delete_metadef_namespace(self, metadef_namespace, ignore_missing=True):
+        """Delete a metadef namespace
+
+        :param metadef_namespace: The value can be either the name of a metadef
+            namespace or a
+            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
+            instance.
+        :param bool ignore_missing: When set to ``False``,
+            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
+            the metadef namespace does not exist.
+        :returns: ``None``
+        """
+        self._delete(
+            _metadef_namespace.MetadefNamespace,
+            metadef_namespace,
+            ignore_missing=ignore_missing,
+        )
+
+    # NOTE(stephenfin): There is no 'find_metadef_namespace' since namespaces
+    # are identified by the namespace name, not an arbitrary UUID, meaning
+    # 'find_metadef_namespace' would be identical to 'get_metadef_namespace'
+
+    def get_metadef_namespace(self, metadef_namespace):
+        """Get a single metadef namespace
+
+        :param metadef_namespace: Either the name of a metadef namespace or an
+            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
+            instance.
+
+        :returns: One
+            :class:`~~openstack.image.v2.metadef_namespace.MetadefNamespace`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+            resource can be found.
+        """
+        return self._get(
+            _metadef_namespace.MetadefNamespace,
+            metadef_namespace,
+        )
+
+    def metadef_namespaces(self, **query):
+        """Return a generator of metadef namespaces
+
+        :returns: A generator object of metadef namespaces
+        :rtype: :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._list(_metadef_namespace.MetadefNamespace, **query)
+
+    def update_metadef_namespace(self, metadef_namespace, **attrs):
+        """Update a server
+
+        :param metadef_namespace: Either the name of a metadef namespace or an
+            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
+            instance.
+        :param attrs: The attributes to update on the metadef namespace
+            represented by ``metadef_namespace``.
+
+        :returns: The updated metadef namespace
+        :rtype: :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
+        """
+        # rather annoyingly, Glance insists on us providing the 'namespace'
+        # argument, even if we're not changing it...
+        if 'namespace' not in attrs:
+            attrs['namespace'] = resource.Resource._get_id(metadef_namespace)
+
+        return self._update(
+            _metadef_namespace.MetadefNamespace,
+            metadef_namespace,
+            **attrs,
+        )
+
+    # ====== SCHEMAS ======
     def get_images_schema(self):
         """Get images schema
 
         :returns: One :class:`~openstack.image.v2.schema.Schema`
         :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
+            when no resource can be found.
         """
         return self._get(_schema.Schema, requires_id=False,
                          base_path='/schemas/images')
@@ -592,7 +773,7 @@ class Proxy(_base_proxy.BaseImageProxy):
 
         :returns: One :class:`~openstack.image.v2.schema.Schema`
         :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
+            when no resource can be found.
         """
         return self._get(_schema.Schema, requires_id=False,
                          base_path='/schemas/image')
@@ -602,7 +783,7 @@ class Proxy(_base_proxy.BaseImageProxy):
 
         :returns: One :class:`~openstack.image.v2.schema.Schema`
         :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
+            when no resource can be found.
         """
         return self._get(_schema.Schema, requires_id=False,
                          base_path='/schemas/members')
@@ -612,16 +793,137 @@ class Proxy(_base_proxy.BaseImageProxy):
 
         :returns: One :class:`~openstack.image.v2.schema.Schema`
         :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
+            when no resource can be found.
         """
         return self._get(_schema.Schema, requires_id=False,
                          base_path='/schemas/member')
 
+    def get_tasks_schema(self):
+        """Get image tasks schema
+
+        :returns: One :class:`~openstack.image.v2.schema.Schema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_schema.Schema, requires_id=False,
+                         base_path='/schemas/tasks')
+
+    def get_task_schema(self):
+        """Get image task schema
+
+        :returns: One :class:`~openstack.image.v2.schema.Schema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_schema.Schema, requires_id=False,
+                         base_path='/schemas/task')
+
+    def get_metadef_namespace_schema(self):
+        """Get metadata definition namespace schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/namespace')
+
+    def get_metadef_namespaces_schema(self):
+        """Get metadata definition namespaces schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/namespaces')
+
+    def get_metadef_resource_type_schema(self):
+        """Get metadata definition resource type association schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/resource_type')
+
+    def get_metadef_resource_types_schema(self):
+        """Get metadata definition resource type associations schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/resource_types')
+
+    def get_metadef_object_schema(self):
+        """Get metadata definition object schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/object')
+
+    def get_metadef_objects_schema(self):
+        """Get metadata definition objects schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/objects')
+
+    def get_metadef_property_schema(self):
+        """Get metadata definition property schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/property')
+
+    def get_metadef_properties_schema(self):
+        """Get metadata definition properties schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/properties')
+
+    def get_metadef_tag_schema(self):
+        """Get metadata definition tag schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/tag')
+
+    def get_metadef_tags_schema(self):
+        """Get metadata definition tags schema
+
+        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
+        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+            when no resource can be found.
+        """
+        return self._get(_metadef_schema.MetadefSchema, requires_id=False,
+                         base_path='/schemas/metadefs/tags')
+
+    # ====== TASKS ======
     def tasks(self, **query):
         """Return a generator of tasks
 
         :param kwargs query: Optional query parameters to be sent to limit
-                             the resources being returned.
+            the resources being returned.
 
         :returns: A generator of task objects
         :rtype: :class:`~openstack.image.v2.task.Task`
@@ -632,11 +934,11 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Get task details
 
         :param task: The value can be the ID of a task or a
-                     :class:`~openstack.image.v2.task.Task` instance.
+            :class:`~openstack.image.v2.task.Task` instance.
 
         :returns: One :class:`~openstack.image.v2.task.Task`
         :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
+            when no resource can be found.
         """
         return self._get(_task.Task, task)
 
@@ -657,22 +959,22 @@ class Proxy(_base_proxy.BaseImageProxy):
         """Wait for a task to be in a particular status.
 
         :param task: The resource to wait on to reach the specified status.
-                    The resource must have a ``status`` attribute.
+            The resource must have a ``status`` attribute.
         :type resource: A :class:`~openstack.resource.Resource` object.
         :param status: Desired status.
         :param failures: Statuses that would be interpreted as failures.
         :type failures: :py:class:`list`
         :param interval: Number of seconds to wait before to consecutive
-                         checks. Default to 2.
+            checks. Default to 2.
         :param wait: Maximum number of seconds to wait before the change.
-                     Default to 120.
+            Default to 120.
         :returns: The resource is returned on success.
         :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
-                 to the desired status failed to occur in specified seconds.
+            to the desired status failed to occur in specified seconds.
         :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
-                 has transited to one of the failure statuses.
+            has transited to one of the failure statuses.
         :raises: :class:`~AttributeError` if the resource does not have a
-                ``status`` attribute.
+            ``status`` attribute.
         """
         if failures is None:
             failures = ['failure']
@@ -714,26 +1016,7 @@ class Proxy(_base_proxy.BaseImageProxy):
             self.log.debug('Still waiting for resource %s to reach state %s, '
                            'current state is %s', name, status, new_status)
 
-    def get_tasks_schema(self):
-        """Get image tasks schema
-
-        :returns: One :class:`~openstack.image.v2.schema.Schema`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
-        """
-        return self._get(_schema.Schema, requires_id=False,
-                         base_path='/schemas/tasks')
-
-    def get_task_schema(self):
-        """Get image task schema
-
-        :returns: One :class:`~openstack.image.v2.schema.Schema`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
-        """
-        return self._get(_schema.Schema, requires_id=False,
-                         base_path='/schemas/task')
-
+    # ====== STORES ======
     def stores(self, **query):
         """Return a generator of supported image stores
 
@@ -742,11 +1025,28 @@ class Proxy(_base_proxy.BaseImageProxy):
         """
         return self._list(_si.Store, **query)
 
+    # ====== IMPORTS ======
     def get_import_info(self):
         """Get a info about image constraints
 
         :returns: One :class:`~openstack.image.v2.service_info.Import`
         :raises: :class:`~openstack.exceptions.ResourceNotFound`
-                 when no resource can be found.
+            when no resource can be found.
         """
         return self._get(_si.Import, require_id=False)
+
+    # ====== UTILS ======
+    def wait_for_delete(self, res, interval=2, wait=120):
+        """Wait for a resource to be deleted.
+
+        :param res: The resource to wait on to be deleted.
+        :type resource: A :class:`~openstack.resource.Resource` object.
+        :param interval: Number of seconds to wait before to consecutive
+            checks. Default to 2.
+        :param wait: Maximum number of seconds to wait before the change.
+            Default to 120.
+        :returns: The resource is returned on success.
+        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
+            to delete failed to occur in the specified seconds.
+        """
+        return resource.wait_for_delete(self, res, interval, wait)
